from __future__ import annotations

import json
from typing import Iterator, List, Optional

from workspaces.capability.enums import CapabilityStepType
from workspaces.capability.helpers_interfaces import (
    CapabilitySequenceIF,
    CapabilityStepIF,
    ParameterIF,
)
from workspaces.capability.schema_interfaces import CapabilityExecutionIF
from workspaces.capability.services.interfaces import CapabilityEngineIF


class CapabilityStep(CapabilityStepIF):
    """
    Class that represents a step in a capability sequence. A step is of a certain type and
    certain step types can have a value
    """

    TYPES = {}

    def __init__(
        self,
        step_type: CapabilityStepType,
        step_value: Optional[str],
        step_args: Optional[str],
    ):
        self.step_type = step_type
        self.step_value = step_value
        self.step_args = step_args

    @classmethod
    def from_str(cls, step_string: str):
        """
        Create CapabilityStep from string containing space-separated type, value, and args

        :param step_string: String of capability step, e.g. "PrepareAndRunWorkflow null"
        :return: CapabilityStep of given string
        """
        step_list = step_string.split(" ")
        step_type = CapabilityStepType.from_string(step_list[0])
        if step_type is CapabilityStepType.PrepareAndRunWorkflow:
            step_value = step_list[1]
            step_args = step_list[2] if len(step_list) == 3 else None
        else:
            # Any other step type
            step_value = step_list[1] if len(step_list) == 2 else None
            step_args = None

        return cls.TYPES[step_type](step_type, step_value, step_args)

    def __str__(self):
        string = f"{self.step_type.name}"
        if self.step_value:
            string += f" {self.step_value}"
        elif self.step_args:
            string += f" {self.step_args}"
        return string


class CapabilitySequence(CapabilitySequenceIF):
    """
    Class that represents a capability sequence with a list of steps
    """

    def __init__(self, steps: List[CapabilityStep]):
        self.steps = steps

    @classmethod
    def from_str(cls, sequence_str: str):
        """
        Create CapabilitySequence from comma-separated str of capability steps

        :param sequence_str: String sequence of steps
        :return: CapabilitySequence of given steps
        """
        steps = []
        for step in sequence_str.split("\n"):
            steps.append(CapabilityStep.from_str(step))

        return cls(steps)

    def __iter__(self) -> Iterator:
        """
        Allows CapabilitySequence objects to be iterated over

        :return: Iterator object
        """
        return iter(self.steps)

    def __len__(self) -> int:
        """
        Allows len() to be called on CapabilitySequence objects

        :return: Number of capability steps
        """
        return len(self.steps)

    def __getitem__(self, step_number: int) -> CapabilityStep:
        return self.steps[step_number]

    def __str__(self) -> str:
        """
        Provides a string representation of the object

        :return: String representation
        """
        str_steps = [str(step) for step in self]
        return ",".join(str_steps)


class PrepareAndRunWorkflow(CapabilityStep):
    def execute(self, engine: CapabilityEngineIF, execution: CapabilityExecutionIF):
        workflow_name = self.step_value
        parameters = execution.capability_request.parameters
        # FIXME: Add support for files
        files = []
        if parameters is not None:
            # convert to json string because json cares about single vs. double quotes.
            # DO NOT TAKE THIS OUT! Python will yell at you.
            parameters = json.dumps(parameters)
            workflow_args = json.loads(parameters)
        engine.submit_workflow_request(execution.id, workflow_name, workflow_args, files)


class AwaitQa(CapabilityStep):
    def execute(self, engine: CapabilityEngineIF, execution: CapabilityExecutionIF):
        pass


class AwaitWorkflow(CapabilityStep):
    def execute(self, engine: CapabilityEngineIF, execution: CapabilityExecutionIF):
        pass


class AwaitProduct(CapabilityStep):
    def execute(self, engine: CapabilityEngineIF, execution: CapabilityExecutionIF):
        pass


class AwaitParameter(CapabilityStep):
    def execute(self, engine: CapabilityEngineIF, execution: CapabilityExecutionIF):
        pass


class AwaitLargeAllocationApproval(CapabilityStep):
    def execute(self, engine: CapabilityEngineIF, execution: CapabilityExecutionIF):
        pass


CapabilityStep.TYPES = {
    CapabilityStepType.PrepareAndRunWorkflow: PrepareAndRunWorkflow,
    CapabilityStepType.AwaitLargeAllocApproval: AwaitLargeAllocationApproval,
    CapabilityStepType.AwaitParameter: AwaitParameter,
    CapabilityStepType.AwaitProduct: AwaitProduct,
    CapabilityStepType.AwaitQA: AwaitQa,
    CapabilityStepType.AwaitWorkflow: AwaitWorkflow,
}


class Parameter(ParameterIF):
    def json(self) -> str:
        pass
